@@ -40,6 +40,7 @@
 #include <camel/camel-data-wrapper.h>
 #include <camel/camel-multipart.h>
 #include <camel/camel-private.h>
+#include <camel/camel-session.h>
 #include <camel/camel-stream-buffer.h>
 #include <camel/camel-stream-mem.h>
 #include <camel/camel-debug.h>
@@ -207,7 +208,6 @@ fetch_items_cb (FetchItemsCallbackData *item_data, gpointer data)
 	struct timeval fi_data_mod_time = { 0 };
 	guint32 j = 0;
 	NTTIME ntdate;
-	gchar *from_name, *from_email;
 
 	MapiItem *item = g_new0(MapiItem , 1);
 
@@ -333,7 +333,7 @@ mapi_update_cache (CamelFolder *folder, GSList *list, CamelFolderChangeInfo **ch
 	CamelFolderChangeInfo *changes = NULL;
 	gboolean exists = FALSE;
 	GString *str = g_string_new (NULL);
-	const gchar *folder_id = NULL;
+	const gchar *from_email, *folder_id = NULL;
 	GSList *item_list = list;
 	int total_items = g_slist_length (item_list), i=0;
 
@@ -346,12 +346,12 @@ mapi_update_cache (CamelFolder *folder, GSList *list, CamelFolderChangeInfo **ch
 		return;
 	}
 
-	camel_operation_start (NULL, _("Fetching summary information for new messages in %s"), folder->name);
+	camel_operation_start (NULL, _("Updating local summary cache for new messages in %s"), folder->name);
 
 	for ( ; item_list != NULL ; item_list = g_slist_next (item_list) ) {
 		MapiItem *temp_item ;
 		MapiItem *item;
-		gchar *msg_uid, *to = NULL, *from = NULL;
+		gchar *msg_uid, *to = NULL, *from = NULL, *cc = NULL;
 		guint64 id;
 
 		exists = FALSE;
@@ -389,7 +389,7 @@ mapi_update_cache (CamelFolder *folder, GSList *list, CamelFolderChangeInfo **ch
 
 		if (!exists) {
  			GSList *l = NULL;
- 			guint32 i =0;
+ 			guint32 count_to = 0, count_cc =0;
 
 			mi->info.uid = exchange_mapi_util_mapi_ids_to_uid(item->fid, item->mid);
 			mi->info.subject = camel_pstring_strdup(item->header.subject);
@@ -397,9 +397,9 @@ mapi_update_cache (CamelFolder *folder, GSList *list, CamelFolderChangeInfo **ch
 			mi->info.size = (guint32) item->header.size;
 
  			for (l = item->recipients; l; l=l->next) {
- 				char *display_name, *name = NULL, *formatted_id;
- 				uint32_t rcpt_type = MAPI_TO;
- 				uint32_t *type = NULL; 
+ 				gchar *formatted_id = NULL;
+				const char *name, *display_name;
+ 				guint32 *type = NULL;
  				struct SRow aRow;
  				ExchangeMAPIRecipient *recip = (ExchangeMAPIRecipient *)(l->data);
  				
@@ -411,55 +411,77 @@ mapi_update_cache (CamelFolder *folder, GSList *list, CamelFolderChangeInfo **ch
  				aRow.ulAdrEntryPad = 0;
  				aRow.cValues = recip->out.all_cValues;
  				aRow.lpProps = recip->out.all_lpProps;
- 				
+
  				type = (uint32_t *) find_SPropValue_data(&aRow, PR_RECIPIENT_TYPE);
- 
- 				if (*type == MAPI_TO) {
- 					/*Name is probably available in one of these props.*/
+
+				if (type) {
  					name = (const char *) find_SPropValue_data(&aRow, PR_DISPLAY_NAME);
  					name = name ? name : (const char *) find_SPropValue_data(&aRow, PR_RECIPIENT_DISPLAY_NAME);
  					name = name ? name : (const char *) find_SPropValue_data(&aRow, 
  												 PR_RECIPIENT_DISPLAY_NAME_UNICODE);
  					name = name ? name : (const char *) find_SPropValue_data(&aRow, 
  												 PR_7BIT_DISPLAY_NAME_UNICODE);
- 					display_name = name ? name : g_strdup (recip->email_id);
+ 					display_name = name ? name : recip->email_id;
  					formatted_id = camel_internet_address_format_address(display_name, recip->email_id);
- 
- 					/* hmm */
- 					if (i)
- 						to = g_string_append (to, ", ");
- 
- 					g_string_append_printf (to, "%s", formatted_id);
- 					i++;
- 				}
- 
- 				/*TODO : from ? */
+
+					switch (*type) {
+					case MAPI_TO:
+						if (count_to) {
+							to = g_strconcat (to, ", ", formatted_id, NULL);
+							g_free (formatted_id);
+						} else
+							to = formatted_id;
+						count_to ++;
+						break;
+
+					case MAPI_CC:
+						if (count_cc) {
+							cc = g_strconcat (cc, ", ", formatted_id, NULL);
+							g_free (formatted_id);
+						} else
+							cc = formatted_id;
+						count_cc ++;
+						break;
+
+					default:
+						continue;
+					}
+				}
  			}
  			
  			if ((item->header.from_type != NULL) && !g_utf8_collate (item->header.from_type, "EX")) {
  				CAMEL_SERVICE_REC_LOCK (mapi_store, connect_lock);
- 				item->header.from_email = exchange_mapi_util_ex_to_smtp (item->header.from_email);
+				from_email = exchange_mapi_util_ex_to_smtp (item->header.from_email);
  				CAMEL_SERVICE_REC_UNLOCK (mapi_store, connect_lock);
+
+ 				g_free (item->header.from_email);
+				item->header.from_email = g_strdup (from_email);				
  			}
  
  			item->header.from_email = item->header.from_email ? 
  				item->header.from_email : item->header.from;
- 
- 			from = camel_internet_address_format_address (item->header.from, 
- 								      item->header.from_email);
- 			mi->info.from = camel_pstring_strdup (from);
- 			mi->info.to = camel_pstring_strdup (to);
 
+			if (item->header.from_email) {
+				from = camel_internet_address_format_address (item->header.from, 
+ 									      item->header.from_email);
+ 				mi->info.from = camel_pstring_strdup (from);
+			} else
+				mi->info.from = NULL;
 
+			/* Fallback */
+ 			mi->info.to = to ? camel_pstring_strdup (to) : g_strdup (item->header.to);
+ 			mi->info.cc = to ? camel_pstring_strdup (cc) : g_strdup (item->header.cc);
 		}
 
 		if (exists) {
 			camel_folder_change_info_change_uid (changes, mi->info.uid);
 			camel_message_info_free (pmi);
 		} else {
+			CAMEL_MAPI_FOLDER_REC_LOCK (folder, cache_lock);
 			camel_folder_summary_add (folder->summary,(CamelMessageInfo *)mi);
 			camel_folder_change_info_add_uid (changes, mi->info.uid);
 			camel_folder_change_info_recent_uid (changes, mi->info.uid);
+			CAMEL_MAPI_FOLDER_REC_UNLOCK (folder, cache_lock);
 		}
 
 		/********************* Summary ends *************************/
@@ -490,6 +512,137 @@ mapi_utils_do_flags_diff (flags_diff_t *diff, guint32 old, guint32 _new)
 	diff->bits = _new & diff->changed;
 }
 
+
+struct mapi_update_deleted_msg {
+	CamelSessionThreadMsg msg;
+
+	CamelFolder *folder;
+	mapi_id_t folder_id;
+};
+
+static gboolean
+deleted_items_sync_cb (FetchItemsCallbackData *item_data, gpointer data)
+{
+	GSList **uid_list = (GSList **) data;
+	gchar *msg_uid = exchange_mapi_util_mapi_ids_to_uid (item_data->fid, 
+							     item_data->mid);
+
+	*uid_list = g_slist_prepend (*uid_list, msg_uid);
+
+	/* Progress update */
+	if (item_data->total > 0)
+		camel_operation_progress (NULL, (item_data->index * 100)/item_data->total);
+
+	/* Check if we have to stop */
+	if (camel_operation_cancel_check(NULL))
+		return FALSE;
+
+	return TRUE;
+}
+
+static void
+mapi_sync_deleted (CamelSession *session, CamelSessionThreadMsg *msg)
+{
+	struct mapi_update_deleted_msg *m = (struct mapi_update_deleted_msg *)msg;
+
+	CamelMapiStore *mapi_store = CAMEL_MAPI_STORE (m->folder->parent_store);
+	CamelMapiFolder *mapi_folder = CAMEL_MAPI_FOLDER (m->folder);
+	CamelFolderChangeInfo *changes = NULL;
+	CamelMessageInfo *info = NULL;
+
+	guint32 index, count, options = 0;
+	GSList *server_uid_list = NULL;
+	const gchar *uid = NULL;
+	gchar *tmp = NULL;
+
+	/* Currently we don't have simple wrapper over getprops.*/
+	const guint32 prop_list[] = { PR_LAST_MODIFICATION_TIME };
+	
+	if (((CamelOfflineStore *) mapi_store)->state == CAMEL_OFFLINE_STORE_NETWORK_UNAVAIL ||
+			((CamelService *)mapi_store)->status == CAMEL_SERVICE_DISCONNECTED) {
+	
+		return;
+	}
+
+	camel_operation_start (NULL, _("Retriving message IDs from server for %s"), m->folder->name);
+
+	CAMEL_SERVICE_REC_LOCK (mapi_store, connect_lock);
+
+	/*Get the UID list from server.*/
+	exchange_mapi_connection_fetch_items  (m->folder_id, NULL, NULL,
+					       prop_list, G_N_ELEMENTS (prop_list),
+					       NULL, NULL,
+					       deleted_items_sync_cb, &server_uid_list,
+					       options);
+
+	camel_operation_end (NULL);
+
+	CAMEL_SERVICE_REC_UNLOCK (mapi_store, connect_lock);
+
+	/* Check if we have to stop */
+	if (camel_operation_cancel_check(NULL))
+		return;
+
+	changes = camel_folder_change_info_new ();
+
+	count = camel_folder_summary_count (m->folder->summary);
+	camel_operation_start (NULL, _("Removing deleted messages from cache in %s"), m->folder->name);
+
+	/* Iterate over cache and check if the UID is in server*/
+	for (index = 0; index < count; index++) {
+		/*FIXME :Any other list available ???*/
+		info = camel_folder_summary_index (m->folder->summary, index);
+		if (!info) continue; /*This is bad. *Should* not happen*/
+
+		uid = camel_message_info_uid (info);
+
+		if (server_uid_list) {
+			/* TODO : Find a better way and avoid this linear search */
+			tmp = g_slist_find_custom (server_uid_list, (const gchar *)uid, 
+						  (GCompareFunc) g_strcmp0);
+		}
+
+		/* If it is not in server list, clean our cache */
+		if (!tmp && uid) {
+			CAMEL_MAPI_FOLDER_REC_LOCK (m->folder, cache_lock);
+			camel_folder_summary_remove_uid (m->folder->summary, uid);
+			camel_data_cache_remove (mapi_folder->cache, "cache", uid, NULL);
+			camel_folder_change_info_remove_uid (changes, uid);
+			CAMEL_MAPI_FOLDER_REC_UNLOCK (m->folder, cache_lock);
+		}
+
+		/* Progress update */
+		camel_operation_progress (NULL, (index * 100)/count); /* ;-) */
+
+		/* Check if we have to stop */
+		if (camel_operation_cancel_check(NULL))
+			return;
+	}
+
+	camel_operation_end (NULL);
+
+	camel_object_trigger_event (m->folder, "folder_changed", changes);
+	camel_folder_change_info_free (changes);
+
+	/* Discard server uid list */
+	g_slist_foreach (server_uid_list, (GFunc) g_free, NULL);
+	g_slist_free (server_uid_list);
+}
+
+static void
+mapi_sync_deleted_free (CamelSession *session, CamelSessionThreadMsg *msg)
+{
+	struct mapi_update_deleted_msg *m = (struct mapi_update_deleted_msg *)msg;
+
+	camel_object_unref (m->folder);
+	/* camel_session_thread_msg_free (session, &m->msg); */
+}
+
+static CamelSessionThreadOps deleted_items_sync_ops = {
+	mapi_sync_deleted,
+	mapi_sync_deleted_free,
+};
+
 static void
 mapi_sync (CamelFolder *folder, gboolean expunge, CamelException *ex)
 {
@@ -502,7 +655,7 @@ mapi_sync (CamelFolder *folder, gboolean expunge, CamelException *ex)
 	flags_diff_t diff, unset_flags;
 	const char *folder_id;
 	mapi_id_t fid, deleted_items_fid;
-	int count, i;
+	gint count, i;
 	guint32 options =0;
 
 	GSList *deleted_items, *deleted_head;
@@ -565,17 +718,14 @@ mapi_sync (CamelFolder *folder, gboolean expunge, CamelException *ex)
 						deleted_head = NULL;
 						deleted_head = deleted_items = g_slist_prepend (deleted_items, mid);
 					}
-
-					CAMEL_SERVICE_REC_LOCK (mapi_store, connect_lock);
-
-					}
-				}
-				
-				if (diff.bits & CAMEL_MESSAGE_SEEN) {
-					read_items = g_slist_prepend (read_items, mid);
-				} else if (unset_flags.bits & CAMEL_MESSAGE_SEEN) {
-					unread_items = g_slist_prepend (unread_items, mid);
 				}
+			}
+
+			if (diff.bits & CAMEL_MESSAGE_SEEN) {
+				read_items = g_slist_prepend (read_items, mid);
+			} else if (unset_flags.bits & CAMEL_MESSAGE_SEEN) {
+				unread_items = g_slist_prepend (unread_items, mid);
+			}
 		}
 		camel_message_info_free (info);
 	}
@@ -595,6 +745,7 @@ mapi_sync (CamelFolder *folder, gboolean expunge, CamelException *ex)
 		g_slist_free (read_items);
 	}
 
+	/* Remove messages from server*/
 	if (deleted_items) {
 		CAMEL_SERVICE_REC_LOCK (mapi_store, connect_lock);
 		if (mapi_folder->type & CAMEL_FOLDER_TYPE_TRASH) {
@@ -606,7 +757,8 @@ mapi_sync (CamelFolder *folder, gboolean expunge, CamelException *ex)
 
 		CAMEL_SERVICE_REC_UNLOCK (mapi_store, connect_lock);
 	}
-	/*Remove them from cache*/
+
+	/*Remove messages from local cache*/
 	while (deleted_items) {
 		char* deleted_msg_uid = g_strdup_printf ("%016" G_GINT64_MODIFIER "X%016" G_GINT64_MODIFIER "X", fid, *(mapi_id_t *)deleted_items->data);
 
@@ -618,18 +770,13 @@ mapi_sync (CamelFolder *folder, gboolean expunge, CamelException *ex)
 		deleted_items = g_slist_next (deleted_items);
 	}
 
-
 	if (unread_items) {
-		CAMEL_SERVICE_REC_LOCK (mapi_store, connect_lock);
 		/* TODO */
-		CAMEL_SERVICE_REC_UNLOCK (mapi_store, connect_lock);
 		g_slist_free (unread_items);
 	}
 
 	if (expunge) {
-		CAMEL_SERVICE_REC_LOCK (mapi_store, connect_lock);
 		/* TODO */
-		CAMEL_SERVICE_REC_UNLOCK (mapi_store, connect_lock);
 	}
 
 	CAMEL_SERVICE_REC_LOCK (mapi_store, connect_lock);
@@ -645,12 +792,15 @@ mapi_refresh_folder(CamelFolder *folder, CamelException *ex)
 	CamelMapiStore *mapi_store = CAMEL_MAPI_STORE (folder->parent_store);
 	CamelMapiFolder *mapi_folder = CAMEL_MAPI_FOLDER (folder);
 	CamelMapiSummary *mapi_summary = CAMEL_MAPI_SUMMARY (folder->summary);
+	CamelSession *session = ((CamelService *)folder->parent_store)->session;
+
 	gboolean is_proxy = folder->parent_store->flags & CAMEL_STORE_PROXY;
 	gboolean is_locked = FALSE;
 	gboolean status;
 
 	struct mapi_SRestriction *res = NULL;
 	struct SSortOrderSet *sort = NULL;
+	struct mapi_update_deleted_msg *deleted_items_op_msg;
 	fetch_items_data *fetch_data = g_new0 (fetch_items_data, 1);
 
 	const gchar *folder_id = NULL;
@@ -736,9 +886,11 @@ mapi_refresh_folder(CamelFolder *folder, CamelException *ex)
 			/*BUG : Fix exception string.*/
 			camel_exception_set (ex, CAMEL_EXCEPTION_SERVICE_UNAVAILABLE,
 					     _("This message is not available in offline mode."));
-			goto end2;
+			goto end1;
 		}
 
+		options |= MAPI_OPTIONS_FETCH_RECIPIENTS;
+
 		if (((CamelMapiFolder *)folder)->type & CAMEL_MAPI_FOLDER_PUBLIC)
 			options |= MAPI_OPTIONS_USE_PFSTORE;
 
@@ -753,7 +905,7 @@ mapi_refresh_folder(CamelFolder *folder, CamelException *ex)
 
 		if (!status) {
 			camel_exception_set (ex, CAMEL_EXCEPTION_SERVICE_INVALID, _("Fetching items failed"));
-			goto end2;
+			goto end1;
 		}
 
 		/*Preserve last_modification_time from this fetch for later use with restrictions.*/
@@ -762,17 +914,22 @@ mapi_refresh_folder(CamelFolder *folder, CamelException *ex)
 		camel_folder_summary_touch (folder->summary);
 		mapi_sync_summary (folder, ex);
 
-		camel_object_trigger_event (folder, "folder_changed", fetch_data->changes);
+		CAMEL_SERVICE_REC_UNLOCK (mapi_store, connect_lock);
+		is_locked = FALSE;
 
+		/* Downsync deleted items */
+		deleted_items_op_msg = camel_session_thread_msg_new (session, &deleted_items_sync_ops, 
+							 sizeof (*deleted_items_op_msg));
+		deleted_items_op_msg->folder = folder;
+		deleted_items_op_msg->folder_id = temp_folder_id;
+		camel_object_ref (folder);
+
+		camel_session_thread_queue (session, &deleted_items_op_msg->msg, 0);
+
+		camel_object_trigger_event (folder, "folder_changed", fetch_data->changes);
 		camel_folder_change_info_free (fetch_data->changes);
 	}
 
-
-	CAMEL_SERVICE_REC_UNLOCK (mapi_store, connect_lock);
-	is_locked = FALSE;
-
-end2:
-	//TODO:
 end1:
 	if (is_locked)
 		CAMEL_SERVICE_REC_UNLOCK (mapi_store, connect_lock);
@@ -1046,17 +1203,18 @@ fetch_item_cb (FetchItemsCallbackData *item_data, gpointer data)
 static void
 mapi_msg_set_recipient_list (CamelMimeMessage *msg, MapiItem *item)
 {
-	g_return_if_fail (item->recipients != NULL);
-
 	GSList *l = NULL;
 	CamelInternetAddress *to_addr, *cc_addr, *bcc_addr;
 
+	g_return_if_fail (item->recipients != NULL);
+
 	to_addr = camel_internet_address_new ();
 	cc_addr = camel_internet_address_new ();
 	bcc_addr = camel_internet_address_new ();
 	
 	for (l = item->recipients; l; l=l->next) {
-		char *display_name, *name = NULL;
+		char *display_name;
+		const char *name = NULL;
 		uint32_t rcpt_type = MAPI_TO;
 		uint32_t *type = NULL; 
 		struct SRow aRow;
@@ -1080,7 +1238,7 @@ mapi_msg_set_recipient_list (CamelMimeMessage *msg, MapiItem *item)
 		type = (uint32_t *) find_SPropValue_data(&aRow, PR_RECIPIENT_TYPE);
 		
 		/*Fallbacks. Not good*/
-		display_name = name ? name : g_strdup (recip->email_id);
+		display_name = name ? g_strdup (name) : g_strdup (recip->email_id);
 		rcpt_type = (type ? *type : MAPI_TO);
 		
 		switch (rcpt_type) {
@@ -1110,6 +1268,7 @@ static void
 mapi_populate_details_from_item (CamelFolder *folder, CamelMimeMessage *msg, MapiItem *item)
 {
 	char *temp_str = NULL;
+	const char *from_email;
 	time_t recieved_time;
 	CamelInternetAddress *addr = NULL;
 	CamelMapiStore *mapi_store = CAMEL_MAPI_STORE(folder->parent_store);
@@ -1129,8 +1288,10 @@ mapi_populate_details_from_item (CamelFolder *folder, CamelMimeMessage *msg, Map
 	if (item->header.from) {
 		if ((item->header.from_type != NULL) && !g_utf8_collate (item->header.from_type, "EX")) {
 			CAMEL_SERVICE_REC_LOCK (mapi_store, connect_lock);
-			item->header.from_email = exchange_mapi_util_ex_to_smtp (item->header.from_email);
+			from_email = exchange_mapi_util_ex_to_smtp (item->header.from_email);
 			CAMEL_SERVICE_REC_UNLOCK (mapi_store, connect_lock);
+			g_free (item->header.from_email);
+			item->header.from_email = g_strdup (from_email);
 		}
 
 		item->header.from_email = item->header.from_email ? 
@@ -1185,7 +1346,7 @@ mapi_folder_item_to_msg( CamelFolder *folder,
 	CamelMultipart *multipart = NULL;
 
 	GSList *attach_list = NULL;
-	int errno;
+	/* int errno; */
 	/* char *body = NULL; */
 	ExchangeMAPIStream *body = NULL;
 	GSList *body_part_list = NULL;
@@ -1284,8 +1445,6 @@ mapi_folder_get_message( CamelFolder *folder, const char *uid, CamelException *e
 	CamelMapiMessageInfo *mi = NULL;
 
 	CamelStream *stream, *cache_stream;
-	int errno;
-
 	mapi_id_t id_folder;
 	mapi_id_t id_message;
 	MapiItem *item = NULL;
@@ -1368,6 +1527,8 @@ mapi_folder_get_message( CamelFolder *folder, const char *uid, CamelException *e
 
 	msg = mapi_folder_item_to_msg (folder, item, ex);
 
+	exchange_mapi_util_free_recipient_list (&item->recipients);
+
 	g_free (item);
 
 	if (!msg) {
@@ -1437,11 +1598,6 @@ mapi_get_message_info(CamelFolder *folder, const char *uid)
 	}
 	msg_info = camel_message_info_new(folder->summary);
 	mi = (CamelMessageInfoBase *)msg_info ;
-	//TODO :
-/* 	oc_message_headers_init(&headers); */
-/* 	oc_thread_connect_lock(); */
-/* 	status = oc_message_headers_get_by_id(&headers, uid); */
-/* 	oc_thread_connect_unlock(); */
 
 	if (headers.subject) mi->subject = (char *)camel_pstring_strdup(headers.subject);
 	if (headers.from) mi->from = (char *)camel_pstring_strdup(headers.from);
@@ -1449,7 +1605,6 @@ mapi_get_message_info(CamelFolder *folder, const char *uid)
 	if (headers.cc) mi->cc = (char *)camel_pstring_strdup(headers.cc);
 	mi->flags = headers.flags;
 
-
 	mi->user_flags = NULL;
 	mi->user_tags = NULL;
 	mi->date_received = 0;
