@@ -71,6 +71,7 @@ mapi_profile_load (const char *profname, const char *password)
 	struct mapi_session *session = NULL;
 	gchar *profpath = NULL;
 	const char *profile = NULL;
+	guint32 debug_log_level = 0;
 
 	d(g_print("\n%s: Entering %s ", G_STRLOC, G_STRFUNC));
 
@@ -90,6 +91,12 @@ mapi_profile_load (const char *profname, const char *password)
 		goto cleanup;
 	}
 
+	if (g_getenv ("MAPI_DEBUG")) {
+		debug_log_level = atoi (g_getenv ("MAPI_DEBUG"));
+		SetMAPIDumpData(TRUE);
+		SetMAPIDebugLevel(debug_log_level);
+	}
+
 	if (profname)
 		profile = profname;
 	else {
@@ -706,16 +713,13 @@ static gboolean
 exchange_mapi_util_get_recipients (mapi_object_t *obj_message, GSList **recip_list)
 {
 	enum MAPISTATUS		retval;
-//	TALLOC_CTX 		*mem_ctx;
 	struct SPropTagArray	proptags;
 	struct SRowSet		rows_recip;
 	uint32_t		i_row_recip;
 	gboolean 		status = TRUE;
 
 	d(g_print("\n%s: Entering %s ", G_STRLOC, G_STRFUNC));
-
-//	mem_ctx = talloc_init ("ExchangeMAPI_GetRecipients");
-
+	
 	/* fetch recipient table */
 	retval = GetRecipientTable(obj_message, &rows_recip, &proptags);
 	if (retval != MAPI_E_SUCCESS) {
@@ -726,12 +730,14 @@ exchange_mapi_util_get_recipients (mapi_object_t *obj_message, GSList **recip_li
 	for (i_row_recip = 0; i_row_recip < rows_recip.cRows; i_row_recip++) {
 		ExchangeMAPIRecipient 	*recipient = g_new0 (ExchangeMAPIRecipient, 1);
 
-		recipient->email_id = (const char *) exchange_mapi_util_find_row_propval (&(rows_recip.aRow[i_row_recip]), PR_SMTP_ADDRESS);
+		recipient->mem_ctx = talloc_init ("ExchangeMAPI_GetRecipients");
+
+		recipient->email_id = talloc_steal (recipient->mem_ctx, (const char *) exchange_mapi_util_find_row_propval (&(rows_recip.aRow[i_row_recip]), PR_SMTP_ADDRESS));
 		/* fallback */
 		if (!recipient->email_id) {
-			const char *addrtype = (const char *) exchange_mapi_util_find_row_propval (&(rows_recip.aRow[i_row_recip]), PR_ADDRTYPE);
+			const char *addrtype = talloc_steal (recipient->mem_ctx, (const char *) exchange_mapi_util_find_row_propval (&(rows_recip.aRow[i_row_recip]), PR_ADDRTYPE));
 			if (addrtype && !g_ascii_strcasecmp(addrtype, "SMTP"))
-				recipient->email_id = (const char *) exchange_mapi_util_find_row_propval (&(rows_recip.aRow[i_row_recip]), PR_EMAIL_ADDRESS);
+				recipient->email_id = talloc_steal (recipient->mem_ctx, (const char *) exchange_mapi_util_find_row_propval (&(rows_recip.aRow[i_row_recip]), PR_EMAIL_ADDRESS));
 		}
 		/* fail */
 		if (!recipient->email_id) {
@@ -739,7 +745,7 @@ exchange_mapi_util_get_recipients (mapi_object_t *obj_message, GSList **recip_li
 			mapidump_SRow (&(rows_recip.aRow[i_row_recip]), " ");
 		}
 
-		recipient->out.all_lpProps = rows_recip.aRow[i_row_recip].lpProps;
+		recipient->out.all_lpProps = talloc_steal ((TALLOC_CTX *)recipient->mem_ctx, rows_recip.aRow[i_row_recip].lpProps);
 		recipient->out.all_cValues = rows_recip.aRow[i_row_recip].cValues;
 
 		*recip_list = g_slist_append (*recip_list, recipient);
@@ -748,7 +754,6 @@ exchange_mapi_util_get_recipients (mapi_object_t *obj_message, GSList **recip_li
 cleanup:
 	if (retval != MAPI_E_SUCCESS)
 		status = FALSE;
-//	talloc_free (mem_ctx);
 
 	d(g_print("\n%s: Leaving %s ", G_STRLOC, G_STRFUNC));
 
@@ -1169,7 +1174,6 @@ exchange_mapi_connection_fetch_items   (mapi_id_t fid,
 				for (k=0; k < prop_count; k++)
 					cast_mapi_SPropValue(&properties_array.lpProps[k], &lpProps[k]);
 
-				MAPIFreeBuffer(lpProps);
 			} else
 				retval = GetPropsAll (&obj_message, &properties_array);
 
@@ -1346,7 +1350,6 @@ exchange_mapi_connection_fetch_item (mapi_id_t fid, mapi_id_t mid,
 		for (k=0; k < prop_count; k++)
 			cast_mapi_SPropValue(&properties_array.lpProps[k], &lpProps[k]);
 
-		MAPIFreeBuffer(lpProps);
 	} else
 		retval = GetPropsAll (&obj_message, &properties_array);
 
@@ -2421,7 +2424,8 @@ cleanup:
 }
 
 static gboolean
-get_child_folders(TALLOC_CTX *mem_ctx, ExchangeMAPIFolderCategory folder_hier, mapi_object_t *parent, mapi_id_t folder_id, GSList **mapi_folders)
+get_child_folders(TALLOC_CTX *mem_ctx, ExchangeMAPIFolderCategory folder_hier, mapi_object_t *parent, 
+		  mapi_id_t folder_id, GSList **mapi_folders, gint32 depth)
 {
 	enum MAPISTATUS		retval;
 	mapi_object_t		obj_folder;
@@ -2435,6 +2439,9 @@ get_child_folders(TALLOC_CTX *mem_ctx, ExchangeMAPIFolderCategory folder_hier, m
 	g_return_val_if_fail (mem_ctx != NULL, FALSE);
 	g_return_val_if_fail (parent != NULL, FALSE);
 
+	/*We reached the depth we wanted.*/
+	if (!depth ) return true;
+
 	mapi_object_init(&obj_folder);
 	mapi_object_init(&obj_table);
 
@@ -2473,6 +2480,8 @@ get_child_folders(TALLOC_CTX *mem_ctx, ExchangeMAPIFolderCategory folder_hier, m
 		goto cleanup;
 	}
 
+	depth--;
+
 	for (i = 0; i < rowset.cRows; i++) {
 		ExchangeMAPIFolder *folder = NULL;
 		gchar *newname = NULL;
@@ -2488,13 +2497,17 @@ get_child_folders(TALLOC_CTX *mem_ctx, ExchangeMAPIFolderCategory folder_hier, m
 			class = IPF_NOTE;
 
 		newname = utf8tolinux (name);
-		g_print("\n|---+ %-15s : (Container class: %s %016" G_GINT64_MODIFIER "X) UnRead : %d Total : %d ", newname, class, *fid, unread ? *unread : 0, total ? *total : 0);
+		g_print("\n|---+ %-15s : (Container class: %s %016" G_GINT64_MODIFIER "X) UnRead : %d Total : %d ", 
+			newname, class, *fid, unread ? *unread : 0, total ? *total : 0);
+
+		folder = exchange_mapi_folder_new (newname, class, folder_hier, *fid, folder_id,
+						   child ? *child : 0, unread ? *unread : 0, total ? *total : 0);
 
-		folder = exchange_mapi_folder_new (newname, class, folder_hier, *fid, folder_id, child ? *child : 0, unread ? *unread : 0, total ? *total : 0);
 		*mapi_folders = g_slist_prepend (*mapi_folders, folder);
 
-		if (child && *child)
-			result = (result && get_child_folders(mem_ctx, folder_hier, &obj_folder, *fid, mapi_folders));
+		if (child && *child && (depth != 0))
+			result = (result && get_child_folders(mem_ctx, folder_hier, &obj_folder, *fid, 
+							      mapi_folders, depth));
 
 		g_free (newname);
 	}
@@ -2714,7 +2727,7 @@ exchange_mapi_get_folders_list (GSList **mapi_folders)
 	*mapi_folders = g_slist_prepend (*mapi_folders, folder);
 
 	/* FIXME: check status of get_child_folders */
-	get_child_folders (mem_ctx, MAPI_PERSONAL_FOLDER, &obj_store, mailbox_id, mapi_folders);
+	get_child_folders (mem_ctx, MAPI_PERSONAL_FOLDER, &obj_store, mailbox_id, mapi_folders, -1);
 
 	g_free(utf8_mailbox_name);
 
@@ -2738,7 +2751,7 @@ cleanup:
 }
 
 gboolean 
-exchange_mapi_get_pf_folders_list (GSList **mapi_folders)
+exchange_mapi_get_pf_folders_list (GSList **mapi_folders, mapi_id_t parent_fid)
 {
 	enum MAPISTATUS 	retval;
 	TALLOC_CTX 		*mem_ctx;
@@ -2746,6 +2759,7 @@ exchange_mapi_get_pf_folders_list (GSList **mapi_folders)
 	gboolean 		result = FALSE;
 	mapi_id_t		mailbox_id;
 	ExchangeMAPIFolder 	*folder;
+	mapi_object_t obj_parent_folder;
 
 	d(g_print("\n%s: Entering %s ", G_STRLOC, G_STRFUNC));
 
@@ -2761,13 +2775,22 @@ exchange_mapi_get_pf_folders_list (GSList **mapi_folders)
 		goto cleanup;
 	}
 
-	/* Prepare the directory listing */
-	retval = GetDefaultPublicFolder(&obj_store, &mailbox_id, olFolderPublicIPMSubtree);
-	if (retval != MAPI_E_SUCCESS) {
-		mapi_errstr("GetDefaultPublicFolder", GetLastError());
-		goto cleanup;
-	}
+	/* Open the folder if parent_fid is given. */
+	if (parent_fid) {
+		mapi_object_init(&obj_parent_folder);
 
+		retval = OpenFolder(&obj_store, parent_fid, &obj_parent_folder);
+		if (retval != MAPI_E_SUCCESS) {
+			mapi_errstr("OpenFolder", GetLastError());
+			goto cleanup;
+		}
+	} else {
+		retval = GetDefaultPublicFolder(&obj_store, &mailbox_id, olFolderPublicIPMSubtree);
+		if (retval != MAPI_E_SUCCESS) {
+			mapi_errstr("GetDefaultPublicFolder", GetLastError());
+			goto cleanup;
+		}
+	}
 	/*  TODO : Localized string */
 	folder = exchange_mapi_folder_new ("All Public Folders", IPF_NOTE, 0, 
 					   mailbox_id, 0, 0, 0 ,0);
@@ -2777,7 +2800,8 @@ exchange_mapi_get_pf_folders_list (GSList **mapi_folders)
 	*mapi_folders = g_slist_prepend (*mapi_folders, folder);
 
 	/* FIXME: check status of get_child_folders */
-	get_child_folders (mem_ctx, MAPI_FAVOURITE_FOLDER, &obj_store, mailbox_id, mapi_folders);
+	get_child_folders (mem_ctx, MAPI_FAVOURITE_FOLDER, parent_fid ? &obj_parent_folder : &obj_store,
+			   parent_fid ? parent_fid : mailbox_id, mapi_folders, 1);
 
 	result = TRUE;
 
